"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.renderBackers = exports.getBackersFromSearch = exports.getBackersFromUsernames = exports.getBackersFromRepositories = exports.getBackers = exports.getGitHubMembersFromOrganizations = exports.getGitHubMembersFromOrganization = exports.getGitHubContributors = exports.getBackersFromGitHubSponsors = exports.getGitHubProfile = exports.getGitHubProfileFromApiUrl = exports.getGitHubOrganization = exports.getGitHubUser = exports.getGitHubRepositoriesFromUsernames = exports.getGitHubRepositoriesFromSearch = exports.getGitHubRepositories = exports.getGitHubRepository = exports.queryGraphQL = exports.queryREST = exports.getCredentialedURL = exports.getURL = exports.removeHeaders = exports.getHeaders = exports.getAuthHeader = exports.getQueryString = exports.redactCredentials = exports.redactSearchParams = exports.removeSearchParams = exports.applySearchParams = exports.getSearchParams = exports.validateCredentials = exports.hasCredentials = exports.getAccessToken = exports.getBackersFromOpenCollective = exports.getBackersFromThanksDev = exports.ThanksDevPlatform = exports.getGitHubSlugFromPackageData = exports.getRepositoryUrlFromPackageData = exports.getRepositoryUrlFromUrlOrGitHubSlug = exports.getRepositoryUrlFromGitHubSlugOrUrl = exports.getRepositoryIssuesUrlFromGitHubSlugOrUrl = exports.getRepositoryWebsiteUrlFromGitHubSlugOrUrl = exports.getGitHubSlugFromUrl = exports.getPackageData = exports.BackersRenderFormat = exports.backerFields = exports.Errlop = exports.PromisePool = exports.Fellow = void 0;
const wait_1 = __importDefault(require("@bevry/wait"));
const native_promise_pool_1 = __importDefault(require("native-promise-pool"));
exports.PromisePool = native_promise_pool_1.default;
const graphql_1 = require("@octokit/graphql");
const js_yaml_1 = require("js-yaml");
const list_1 = require("@bevry/list");
const trim_empty_keys_1 = __importDefault(require("trim-empty-keys"));
const errlop_1 = __importDefault(require("errlop"));
exports.Errlop = errlop_1.default;
const arrange_package_json_1 = __importDefault(require("arrange-package-json"));
const fellow_1 = __importStar(require("fellow"));
exports.Fellow = fellow_1.default;
// defaults
const node_process_1 = require("node:process");
const envCredentials = node_process_1.env;
/** Field names that we consider backers */
exports.backerFields = [
    'author',
    'authors',
    'maintainers',
    'contributors',
    'funders',
    'sponsors',
    'donors',
];
/** A rendering style for {@link Backers} */
var BackersRenderFormat;
(function (BackersRenderFormat) {
    /** Renders as {@link BackersRenderResult.string} with appropriate defaults, intended for `package.json` merging. */
    BackersRenderFormat["string"] = "string";
    /** Renders as {@link BackersRenderResult.text} with appropriate defaults, intended for human-readable outputs.*/
    BackersRenderFormat["text"] = "text";
    /** Renders as {@link BackersRenderResult.markdown} with appropriate defaults, intended for readme outputs. */
    BackersRenderFormat["markdown"] = "markdown";
    /** Renders as {@link BackersRenderResult.html} with appropriate defaults, intended for readme outputs. */
    BackersRenderFormat["html"] = "html";
    /** Same as {@link BackersRenderFormat.string} however merges with {@link BackersRenderOptions.packageData} if it exists. */
    BackersRenderFormat["package"] = "package";
    /** Used for license files, author and authors are an array of strings, using {@link Fellow.toMarkdown} with appropriate defaults. */
    BackersRenderFormat["copyright"] = "copyright";
    /** Used for shoutout.txt files, outputs a string with shoutouts to contributors, funders, and sponsors, using {@link Fellow.toText} with appropriate defaults. */
    BackersRenderFormat["shoutout"] = "shoutout";
    /** Used for initial changelog entry, outputs a string with shoutout to funders and sponsors, using {@link Fellow.toMarkdown} with appropriate defaults. */
    BackersRenderFormat["release"] = "release";
    /** Used for subsequent changelog entries, outputs a string with shoutout to sponsors, using {@link Fellow.toMarkdown} with appropriate defaults. */
    BackersRenderFormat["update"] = "update";
})(BackersRenderFormat || (exports.BackersRenderFormat = BackersRenderFormat = {}));
/** Fetch the package data of a repository. */
async function getPackageData(githubSlug, fallback) {
    const url = `http://raw.github.com/${githubSlug}/master/package.json`;
    const resp = await fetch(url, {});
    if (!resp.ok) {
        const error = new Error(`package.json not found for ${githubSlug}`);
        if (!fallback)
            throw error;
        console.warn(error.stack);
        return fallback;
    }
    try {
        const packageData = await resp.json();
        return packageData;
    }
    catch (err) {
        const error = new errlop_1.default(`failed to parse package.json of ${githubSlug}`, err);
        if (!fallback)
            throw error;
        console.warn(error.stack);
        return fallback;
    }
}
exports.getPackageData = getPackageData;
/**
 * Get the GitHub slug of this GitHub url
 * @example
 * ```
 * bevry/github-api
 * github:bevry/github-api
 * git@github.com:bevry/github-api.git
 * https://github.com/bevry/github-api.git
 * ssh://github.com/bevry/github-api.git
 * git://github.com/user/project.git#commit-ish
 * git+https://github.com/bevry/github-api.git#commit-ish
 * git+ssh://github.com/bevry/github-api.git#commit-ish
 * git+ssh+ssh://github.com/bevry/github-api.git#commit-ish
 * ```
 */
function getGitHubSlugFromUrl(url) {
    // https://docs.npmjs.com/cli/v10/configuring-npm/package-json#repository
    // https://stackoverflow.com/a/31457913/130638
    // Extract the slug a GitHub URL
    const match = url.match(/^(?:github:|git@github[.]com:|(?:(?:git|ssh|https?)[+]?)+:[/][/]github[.]com[/])?([^:/]+[/][^:/]+?)(?:[.]git)?(?:#.*)?$/);
    // Verify and return the slug
    return ((match &&
        match[1] &&
        match[1].split('/').filter(Boolean).length === 2 &&
        match[1]) ||
        null);
}
exports.getGitHubSlugFromUrl = getGitHubSlugFromUrl;
/** Get the GitHub Repository website url of a GitHub slug */
function getRepositoryWebsiteUrlFromGitHubSlugOrUrl(githubSlug = '') {
    const result = getGitHubSlugFromUrl(githubSlug);
    if (!result)
        return result;
    return `https://github.com/${githubSlug}`;
}
exports.getRepositoryWebsiteUrlFromGitHubSlugOrUrl = getRepositoryWebsiteUrlFromGitHubSlugOrUrl;
/** Get the GitHub Repository issues url of a GitHub slug */
function getRepositoryIssuesUrlFromGitHubSlugOrUrl(githubSlug = '') {
    const result = getGitHubSlugFromUrl(githubSlug);
    if (!result)
        return result;
    return `https://github.com/${githubSlug}/issues`;
}
exports.getRepositoryIssuesUrlFromGitHubSlugOrUrl = getRepositoryIssuesUrlFromGitHubSlugOrUrl;
/** Get the GitHub Repository url of a GitHub slug or GitHub url */
function getRepositoryUrlFromGitHubSlugOrUrl(githubSlug) {
    const result = getGitHubSlugFromUrl(githubSlug);
    if (!result)
        return result;
    // this is what npm prefers:
    // https://github.com/bevry/oneday/actions/runs/7338411760/job/19980939564#step:9:22
    return `git+https://github.com/${result}.git`;
}
exports.getRepositoryUrlFromGitHubSlugOrUrl = getRepositoryUrlFromGitHubSlugOrUrl;
/** Get the Repository url of a URL or GitHub slug */
function getRepositoryUrlFromUrlOrGitHubSlug(input) {
    return getRepositoryUrlFromGitHubSlugOrUrl(input) || input || null;
}
exports.getRepositoryUrlFromUrlOrGitHubSlug = getRepositoryUrlFromUrlOrGitHubSlug;
/** Get the Repository url (or shorthand url) from package.json data */
function getRepositoryUrlFromPackageData(packageData) {
    if (typeof packageData.repository === 'string') {
        // modernize if possible
        return getRepositoryUrlFromUrlOrGitHubSlug(packageData.repository);
    }
    else if (packageData.repository &&
        typeof packageData.repository.url === 'string') {
        // modernize if possible
        return getRepositoryUrlFromUrlOrGitHubSlug(packageData.repository && packageData.repository.url);
    }
    else if (typeof packageData.homepage === 'string') {
        // this may or may not be a repository, so if it can be converted then do it
        return getRepositoryUrlFromGitHubSlugOrUrl(packageData.homepage);
    }
    else {
        return null;
    }
}
exports.getRepositoryUrlFromPackageData = getRepositoryUrlFromPackageData;
/** Get the GitHub slug from package.json data  */
function getGitHubSlugFromPackageData(packageData) {
    // extract the url (or shorthand url)
    const url = getRepositoryUrlFromPackageData(packageData);
    // extract the slug
    const githubSlug = url && getGitHubSlugFromUrl(url);
    if (githubSlug)
        return githubSlug;
    // fail if no slug
    throw new errlop_1.default({
        code: 'PACKAGE_DATA_NOT_CONFIGURED_FOR_GITHUB_REPOSITORY',
        message: `Could not determine GitHub slug from package data: ${JSON.stringify(packageData, null, '  ')}`,
    });
}
exports.getGitHubSlugFromPackageData = getGitHubSlugFromPackageData;
/** Fetch the funding data of a repository. */
async function getFundingData(githubSlug, fallback) {
    const url = `http://raw.github.com/${githubSlug}/master/.github/FUNDING.yml`;
    const resp = await fetch(url, {});
    if (!resp.ok) {
        const error = new Error(`.github/FUNDING.yml not found for ${githubSlug}`);
        if (!fallback)
            throw error;
        console.warn(error.stack);
        return fallback;
    }
    try {
        const fundingText = await resp.text();
        const fundingData = (0, js_yaml_1.load)(fundingText);
        return fundingData;
    }
    catch (err) {
        const error = new errlop_1.default(`failed to fetch .github/FUNDING.yml of ${githubSlug}`, err);
        if (!fallback)
            throw error;
        console.warn(error.stack);
        return fallback;
    }
}
// ====================================
// ThanksDev Types
var ThanksDevPlatform;
(function (ThanksDevPlatform) {
    ThanksDevPlatform["GitHub"] = "gh";
    ThanksDevPlatform["GitLab"] = "gl";
})(ThanksDevPlatform || (exports.ThanksDevPlatform = ThanksDevPlatform = {}));
/** Get a Fellow of a ThanksDev member */
function getThanksDevProfile([platform, username, cents,]) {
    return fellow_1.default.ensure({
        thanksdevUrl: `https://thanks.dev/d/${platform}/${username}`,
    });
}
/** Get the {@link Backers} of the from the ThanksDev API */
async function getBackersFromThanksDev(platform, username, opts = {}) {
    const sponsorCentsThreshold = opts.sponsorCentsThreshold ?? 100;
    const donorCentsThreshold = opts.donorCentsThreshold ?? 100;
    let sponsors = [], donors = [];
    // monthly
    try {
        const url = `https://api.thanks.dev/v1/vip/dependee/${platform}/${username}?age=30`; // 30 is average days in a month
        const resp = await fetch(url, {});
        const data = await resp.json();
        sponsors = data.donors.filter(([, , cents]) => cents && (!sponsorCentsThreshold || cents >= sponsorCentsThreshold));
    }
    catch (err) {
        throw new errlop_1.default(`Fetching ThanksDev sponsors failed for: ${platform}/${username}`, err);
    }
    // yearly
    try {
        const url = `https://api.thanks.dev/v1/vip/dependee/${platform}/${username}`;
        const resp = await fetch(url, {});
        const data = await resp.json();
        donors = data.donors.filter(([, , cents]) => cents && (!donorCentsThreshold || cents >= donorCentsThreshold));
    }
    catch (err) {
        throw new errlop_1.default(`Fetching ThanksDev donors failed for: ${platform}/${username}`, err);
    }
    // combine
    return {
        sponsors: fellow_1.default.add(sponsors.map(getThanksDevProfile)),
        donors: fellow_1.default.add(donors.map(getThanksDevProfile)),
    };
}
exports.getBackersFromThanksDev = getBackersFromThanksDev;
/** Return a date instance that is the last month */
function getLastMonth(now = new Date()) {
    const date = new Date();
    date.setMonth(date.getMonth() - 1);
    return date;
}
/** Is the date within the last month */
function isWithinLastMonth(when, lastMonth = getLastMonth()) {
    if (lastMonth instanceof Date)
        lastMonth = lastMonth.getTime();
    if (typeof when === 'string')
        when = new Date(when);
    return when.getTime() >= lastMonth;
}
/** Get a Fellow of an OpenCollective member */
function getOpenCollectiveProfile(member) {
    return fellow_1.default.ensure({
        opencollectiveProfile: member,
        description: member.description,
        githubUrl: member.github,
        name: member.name,
        email: member.email,
        opencollectiveUrl: member.profile,
        twitterUrl: member.twitter,
        websiteUrl: member.website,
    });
}
/** Get the {@link Backers} of the from the OpenCollective API */
async function getBackersFromOpenCollective(username, opts = {}) {
    try {
        const sponsorCentsThreshold = opts.sponsorCentsThreshold ?? 100;
        const donorCentsThreshold = opts.donorCentsThreshold ?? 100;
        const url = `https://opencollective.com/${username}/members.json`;
        const resp = await fetch(url, {});
        const profiles = await resp.json();
        // remove invalid github urls, as they can get outdated perhaps, as opencollective returns https://github.com/mikeumu for @mikeumus which should be https://github.com/mikeumus
        for (const profile of profiles) {
            let github = profile.github;
            if (github && (await (0, fellow_1.fetchNotOk)(github))) {
                const username = profile.profile.replace(/^.+\.com\//, ''); // strip https://opencollective.com/
                github = `https://github.com/${username}`;
                if (username && (await (0, fellow_1.fetchOk)(github))) {
                    profile.github = github;
                }
                else {
                    profile.github = '';
                }
            }
        }
        const lastMonth = getLastMonth().getTime();
        const sponsors = profiles.filter((member) => member.role === 'BACKER' &&
            member.lastTransactionAmount &&
            (!sponsorCentsThreshold ||
                member.lastTransactionAmount * 100 > sponsorCentsThreshold) &&
            isWithinLastMonth(member.lastTransactionAt, lastMonth));
        const donors = profiles.filter((member) => member.role === 'BACKER' &&
            member.totalAmountDonated &&
            (!donorCentsThreshold ||
                member.totalAmountDonated * 100 > donorCentsThreshold));
        return {
            sponsors: fellow_1.default.add(sponsors.map(getOpenCollectiveProfile)),
            donors: fellow_1.default.add(donors.map(getOpenCollectiveProfile)),
        };
    }
    catch (err) {
        throw new errlop_1.default(`Fetching OpenCollective sponsors and donors failed for: ${username}`, err);
    }
}
exports.getBackersFromOpenCollective = getBackersFromOpenCollective;
// ====================================
// GitHub API Methods
/**
 * Get the desired GitHub Access Token from the credentials.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 */
function getAccessToken(credentials = envCredentials) {
    return credentials.GITHUB_ACCESS_TOKEN || credentials.GITHUB_TOKEN || null;
}
exports.getAccessToken = getAccessToken;
/** Check whether or not sufficient GitHub credentials were supplied. */
function hasCredentials(credentials = envCredentials) {
    const accessToken = getAccessToken(credentials);
    const valid = accessToken ||
        (credentials.GITHUB_CLIENT_ID && credentials.GITHUB_CLIENT_SECRET);
    return Boolean(valid);
}
exports.hasCredentials = hasCredentials;
/**
 * Check whether or not sufficient GitHub credentials were supplied.
 * @throws if invalid
 */
function validateCredentials(credentials = envCredentials) {
    if (hasCredentials(credentials))
        return true;
    throw new errlop_1.default({
        code: 'INVALID_GITHUB_AUTH',
        message: 'Insufficient GitHub credentials: provide `GITHUB_ACCESS_TOKEN` or `GITHUB_TOKEN`, or a combination of `GITHUB_CLIENT_ID` and `GITHUB_CLIENT_SECRET`',
    });
}
exports.validateCredentials = validateCredentials;
/**
 * Get the GitHub Authorization Search Params.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 * @param credentials The params to use for the authorization variables.
 * @param params If you wish to set the params on an existing URLSearchParams instance, then provide it here.
 * @returns The search params that you should append to your github API request url.
 * @throws If no valid GitHub Authorization was provided.
 */
function getSearchParams(credentials = envCredentials, params = new URLSearchParams()) {
    validateCredentials(credentials);
    const accessToken = getAccessToken(credentials);
    if (accessToken) {
        params.set('access_token', accessToken);
    }
    else if (credentials.GITHUB_CLIENT_ID && credentials.GITHUB_CLIENT_SECRET) {
        params.set('client_id', credentials.GITHUB_CLIENT_ID);
        params.set('client_secret', credentials.GITHUB_CLIENT_SECRET);
    }
    else {
        throw new errlop_1.default({
            code: 'INVALID_STATE',
            message: 'validateCredentails failed to determine invalid GitHub Credentials',
        });
    }
    return params;
}
exports.getSearchParams = getSearchParams;
function applySearchParams(target, source) {
    if (source) {
        if (source instanceof URLSearchParams) {
            source.forEach((value, key) => target.set(key, value));
        }
        else {
            Object.entries(source).forEach(([key, value]) => target.set(key, value));
        }
    }
}
exports.applySearchParams = applySearchParams;
/**
 * Remove any GitHub Credentials from a URL Search Params instance.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 */
function removeSearchParams(params) {
    params.delete('access_token');
    params.delete('client_id');
    params.delete('client_secret');
    return params;
}
exports.removeSearchParams = removeSearchParams;
/**
 * Redact any GitHub Credentials in their URLSearchParams from a string.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 * @param value The string to redact credentials from.
 * @returns The string with the credentials redacted.
 */
function redactSearchParams(value) {
    return value.replace(/(&?)(access_token|client_id|client_secret)=\w+/gi, '$1$2=REDACTED');
}
exports.redactSearchParams = redactSearchParams;
/** Redact the GitHub Credentials in any form from a string */
function redactCredentials(value, credentials = envCredentials) {
    value = redactSearchParams(value);
    let cred = getAccessToken(credentials);
    if (cred)
        value = value.replaceAll(cred, 'REDACTED');
    cred = credentials.GITHUB_CLIENT_ID;
    if (cred)
        value = value.replaceAll(cred, 'REDACTED');
    cred = credentials.GITHUB_CLIENT_SECRET;
    if (cred)
        value = value.replaceAll(cred, 'REDACTED');
    return value;
}
exports.redactCredentials = redactCredentials;
/**
 * Get the GitHub Authorization as a Query String.
 * You probably want to use {@link getURL} directly, instead of going through this method.
 */
function getQueryString(credentials = envCredentials) {
    return getSearchParams(credentials).toString();
}
exports.getQueryString = getQueryString;
/**
 * Get the GitHub Authorization Header.
 * Use as the `Authorization` header within {@link fetch} calls.
 * You probably want to use {@link getHeaders} or {@link fetch} directly, instead of going through this method.
 * @throws If no valid GitHub Authorization was provided.
 */
function getAuthHeader(credentials = envCredentials) {
    validateCredentials(credentials);
    const accessToken = getAccessToken(credentials);
    if (accessToken) {
        return `token ${accessToken}`;
    }
    else if (credentials.GITHUB_CLIENT_ID && credentials.GITHUB_CLIENT_SECRET) {
        return `Basic ${credentials.GITHUB_CLIENT_ID}:${credentials.GITHUB_CLIENT_SECRET}`;
    }
    else {
        throw new errlop_1.default({
            code: 'INVALID_STATE',
            message: 'validateCredentails failed to determine invalid GitHub Credentials',
        });
    }
}
exports.getAuthHeader = getAuthHeader;
/**
 * Get the headers to attach to the request to the GitHub API.
 * Use as the headers object within {@link fetch} calls.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 */
function getHeaders(credentials = envCredentials, headers = {}) {
    return {
        Accept: 'application/vnd.github.v3+json',
        Authorization: getAuthHeader(credentials),
        ...headers,
    };
}
exports.getHeaders = getHeaders;
/**
 * Remove any GitHub Credentials from a Headers instance.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 */
function removeHeaders(headers) {
    // @ts-ignore
    delete headers.Authorization;
    return headers;
}
exports.removeHeaders = removeHeaders;
/**
 * Get the desired Github API URL, using {@link removeSearchParams}.to ensure there are no credentials.
 * As this URL does not include credentials, use with {@link getAuthHeader} to authorize correctly.
 * Otherwise use {@link getCredentialedURL} to get a credentialed URL.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 * If the credentials property is nullish, then the environment variables are attempted.
 */
function getURL(opts = {}) {
    // fetch url
    const url = new URL(opts.url ||
        (opts.credentials || envCredentials).GITHUB_API_URL ||
        (opts.credentials || envCredentials).GITHUB_API ||
        'https://api.github.com');
    // add user params
    applySearchParams(url.searchParams, opts.searchParams);
    // ensure that there are no credentials in the URL
    removeSearchParams(url.searchParams);
    // add user pathname
    // the convoluted way of doing this is to make sure that with or without / is valid
    // as the GITHUB_API hostname may be something like `https://bevry.me/api/github`
    if (opts.pathname) {
        if (opts.pathname.includes('://'))
            throw new Error(`Received pathname ${opts.pathname} which is not a pathname but a URL`);
        url.pathname = [
            url.pathname.replace(/^[/]+|[/]+$/, ''),
            opts.pathname.replace(/^[/]+|[/]+$/, ''),
        ]
            .filter((i) => i)
            .join('/');
    }
    // return
    return url;
}
exports.getURL = getURL;
/**
 * Get the credentialed GitHub API URL instance.
 * Uses {@link getURL} to get the URL, then uses {@link getSearchParams} to add the credentials.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 * If the credentials property is nullish, then the environment variables are attempted.
 */
function getCredentialedURL(opts = {}) {
    // fetch url
    const url = getURL(opts);
    // add auth params
    getSearchParams(opts.credentials, url.searchParams);
    // return
    return url;
}
exports.getCredentialedURL = getCredentialedURL;
/**
 * Fetches a GitHub REST API response with authentication, parsing, waiting, pooling, paging.
 * If the credentials property is nullish, then the environment variables are attempted.
 * If the user agent is nullish, then it will be set to `"@bevry/github"`
 */
async function queryREST(opts = {}) {
    // defaults
    opts.pool ??= new native_promise_pool_1.default(opts.concurrency);
    const searchParams = new URLSearchParams();
    applySearchParams(searchParams, opts.searchParams);
    if (opts.page != null || opts.pages != null || opts.size != null) {
        opts.page ??= 1;
        opts.pages ??= 0;
        opts.size ??= 100;
    }
    if (opts.page)
        searchParams.set('page', String(opts.page));
    if (opts.size)
        searchParams.set('per_page', String(opts.size));
    // prepare fetch
    const url = getURL({ ...opts, searchParams }); // searchParams consumed here
    const fetchOpts = {
        headers: getHeaders(opts.credentials, {
            'User-Agent': opts.userAgent || '@bevry/github',
            ...(opts.headers || {}),
        }),
        method: opts.method,
        body: opts.body,
    };
    // fetch in the pool
    let error = null;
    let responseREST;
    let responseJSON; // GitHubRESTResponse<T>
    await opts.pool.open(async () => {
        responseREST = await fetch(url, fetchOpts);
        while (responseREST.status === 429) {
            console.warn(`Request to ${url} failed with status 429: Too Many Requests, will try again in a minute...`);
            await (0, wait_1.default)(60 * 1000);
            responseREST = await fetch(url, fetchOpts);
        }
    });
    // check status
    if (responseREST.ok === false) {
        error = new errlop_1.default(`Request to ${url} failed with status ${responseREST.status}: ${responseREST.statusText}`, error);
    }
    // parse
    try {
        responseJSON = await responseREST.json();
    }
    catch (err) {
        responseJSON = {}; // workaround os the rest doesn't fail
        error = new errlop_1.default(`Request to ${url} failed to produce valid JSON`, error);
    }
    // check for errors
    for (const message of [
        responseJSON.message || '',
        ...(responseJSON.errors || []).map((e) => e.message || ''),
    ]) {
        if (message)
            error = new errlop_1.default(message, error);
    }
    // throw error if there is one
    if (error)
        throw error;
    // paging and array checks
    if (opts.page != null) {
        // check
        if (!Array.isArray(responseJSON)) {
            if (Array.isArray(responseJSON.items)) {
                responseJSON = responseJSON.items;
            }
            else {
                throw new Error(`Request to ${url} did not return an array as expected`);
            }
        }
        if (responseJSON.length === 0) {
            return [];
        }
        // another page?
        const within = !opts.pages || opts.page < opts.pages;
        const hasNextPage = responseJSON.length === opts.size && within;
        if (hasNextPage) {
            (0, list_1.append)(responseJSON, await queryREST({
                ...opts,
                page: opts.page + 1,
            }));
        }
    }
    // return
    return responseJSON;
}
exports.queryREST = queryREST;
/**
 * Fetches a GitHub GraphQL API response with authentication, parsing, waiting, pooling. NO PAGING.
 * If the credentials property is nullish, then the environment variables are attempted.
 * If the user agent is nullish, then it will be set to `"@bevry/github"`
 */
async function queryGraphQL(query, opts = {}) {
    // prepare
    opts.pool ??= new native_promise_pool_1.default(opts.concurrency);
    // prepare fetch
    // https://docs.github.com/en/graphql/overview/explorer
    const fetchOpts = {
        headers: getHeaders(opts.credentials, {
            'User-Agent': opts.userAgent || '@bevry/github',
            ...(opts.headers || {}),
        }),
    };
    // fetch in the pool
    let error = null;
    let responseGraphQL;
    await opts.pool.open(async () => {
        responseGraphQL = await (0, graphql_1.graphql)(query, fetchOpts);
        while (responseGraphQL.status === 429) {
            console.warn(`GraphQL returned status code [429 Too Many Requests] will try again in a minute`);
            await (0, wait_1.default)(60 * 1000);
            responseGraphQL = await (0, graphql_1.graphql)(query, fetchOpts);
        }
    });
    // check for errors
    for (const message of [
        responseGraphQL.message || '',
        ...(responseGraphQL.errors || []).map((e) => e.message || ''),
    ]) {
        if (message)
            error = new errlop_1.default(message, error);
    }
    // throw error if there is one
    if (error)
        throw error;
    // return
    return (opts.resultField ? responseGraphQL[opts.resultField] : responseGraphQL);
}
exports.queryGraphQL = queryGraphQL;
// =================================
// GitHub Misc
/** Get the latest commit for a GitHub Repository */
async function getGitHubLatestCommit(githubSlug, opts = {}) {
    // fetch
    const commits = await queryREST({
        ...opts,
        pathname: `repos/${githubSlug}/commits`,
    });
    // unexpected result
    const commit = commits[0] && commits[0].sha;
    if (!commit)
        throw new Error(`GitHub Commit was not present in response for: ${githubSlug}`);
    // success
    return commit;
}
exports.default = getGitHubLatestCommit;
// =================================
// GitHub Repository
/** Get this GitHub Repository */
async function getGitHubRepository(githubSlug, opts = {}) {
    const result = await queryREST({
        ...opts,
        pathname: `repos/${githubSlug}`,
    });
    if (!result || !result.full_name)
        throw new Error(`GitHub Repository was not present in response for: ${githubSlug}`);
    return result;
}
exports.getGitHubRepository = getGitHubRepository;
/** Get the GitHub Repositories of these slugs */
async function getGitHubRepositories(githubSlugs, opts = {}) {
    return await Promise.all(githubSlugs.map((githubSlug) => getGitHubRepository(githubSlug, opts)));
}
exports.getGitHubRepositories = getGitHubRepositories;
/**
 * Get the GitHub Repositories of this search query
 * @param search e.g. `\@bevry language:typescript`
 */
async function getGitHubRepositoriesFromSearch(search, opts = {}) {
    return await queryREST({
        ...opts,
        pathname: `search/repositories`,
        pages: opts.pages ?? 0,
        searchParams: {
            q: search,
        },
    });
}
exports.getGitHubRepositoriesFromSearch = getGitHubRepositoriesFromSearch;
/** Get the GitHub Repositories of these usernames (users, organizations) */
async function getGitHubRepositoriesFromUsernames(usernames, opts = {}) {
    const query = usernames.map((username) => `@${username}`).join('%20');
    return await getGitHubRepositoriesFromSearch(query, opts);
}
exports.getGitHubRepositoriesFromUsernames = getGitHubRepositoriesFromUsernames;
// ====================================
// GitHub Models
/** Get the {@link Fellow} entity of this GitHub User via the GitHub GraphQL API */
async function getGitHubUser(username, opts = {}) {
    try {
        const githubProfile = await queryGraphQL(
        /* GraphQL */ `{
				user(login: "${username}") {
					bio
					company
					email
					isHireable
					location
					login
					name
					url
					websiteUrl
				}
			}`, { ...opts, resultField: 'user' });
        // verify
        if (!githubProfile)
            throw new Error(`No GitHub User was returned for username: ${username}`);
        // add these items
        const fellow = fellow_1.default.ensure({
            githubProfile,
            company: githubProfile.company,
            description: githubProfile.bio,
            email: githubProfile.email,
            githubUrl: githubProfile.url,
            githubUsername: githubProfile.login,
            hireable: githubProfile.isHireable,
            location: githubProfile.location,
            name: githubProfile.name,
            websiteUrl: githubProfile.websiteUrl,
        });
        return fellow;
    }
    catch (err) {
        throw new errlop_1.default(`Fetching GitHub User failed for username: ${username}`, err);
    }
}
exports.getGitHubUser = getGitHubUser;
/** Get the {@link Fellow} entity of this GitHub Organization via the GitHub GraphQL API */
async function getGitHubOrganization(username, opts = {}) {
    // fetch
    // https://docs.github.com/en/graphql/overview/explorer
    try {
        const githubProfile = await queryGraphQL(
        /* GraphQL */ `{
				organization(login: "${username}") {
					description
					email
					location
					login
					name
					url
					websiteUrl
				}
			}`, { ...opts, resultField: 'organization' });
        // verify
        if (!githubProfile)
            throw new Error(`No GitHub Organization was returned for username: ${username}`);
        // add these items
        const fellow = fellow_1.default.ensure({
            githubProfile,
            description: githubProfile.description,
            email: githubProfile.email,
            githubUrl: githubProfile.url,
            githubUsername: githubProfile.login,
            location: githubProfile.location,
            name: githubProfile.name,
            websiteUrl: githubProfile.websiteUrl,
        });
        return fellow;
    }
    catch (err) {
        throw new errlop_1.default(`Fetching GitHub Organization failed for username: ${username}`, err);
    }
}
exports.getGitHubOrganization = getGitHubOrganization;
/** Get the {@link Fellow} entity of this GitHub User/Organisation via the GitHub REST API */
async function getGitHubProfileFromApiUrl(url, opts = {}) {
    if (url.includes('api.github.com') === false)
        throw new Error(`Cannot fetch the GitHub Profile for non-API URL: ${url}`);
    const githubProfile = await queryREST({
        ...opts,
        url,
    });
    const fellow = fellow_1.default.ensure({
        githubProfile,
        company: githubProfile.company,
        description: githubProfile.bio,
        email: githubProfile.email,
        githubUrl: githubProfile.html_url,
        githubUsername: githubProfile.login,
        hireable: githubProfile.hireable,
        homepage: githubProfile.blog,
        location: githubProfile.location,
        name: githubProfile.name,
    });
    return fellow;
}
exports.getGitHubProfileFromApiUrl = getGitHubProfileFromApiUrl;
/** Get the {@link Fellow} entity of this GitHub User/Organisation via the GitHub API */
async function getGitHubProfile(username, opts = {}) {
    // await necessary for try catch to catch promise failures
    if (username.includes('api.github.com')) {
        return await getGitHubProfileFromApiUrl(username, opts);
    }
    try {
        return await getGitHubUser(username, opts);
    }
    catch (userError) {
        try {
            return await getGitHubOrganization(username, opts);
        }
        catch (organizationError) {
            throw new errlop_1.default(`Failed to fetch the GitHub Profile for the username: ${username}`, new errlop_1.default(organizationError, userError));
        }
    }
}
exports.getGitHubProfile = getGitHubProfile;
/** Get the {@link Fellow} entities for the GitHub Sponsors of this GitHub Profile via the GitHub Sponsors API */
async function getBackersFromGitHubSponsors(username, opts = {}) {
    try {
        // fetch
        // https://docs.github.com/en/graphql/reference/interfaces#sponsorable
        // https://stackoverflow.com/a/65272597/130638
        // https://docs.github.com/en/graphql/reference/objects#sponsorconnection
        // https://docs.github.com/en/graphql/overview/explorer
        // @todo no idea how to fetch the cents the user gave in the past month
        // only possible to fetch your prior sponsorships as a sponsor: https://github.com/orgs/community/discussions/38393
        // not possible to fetch your prior sponsors as someone who receives sponsorship
        // although perhaps https://github.com/orgs/community/discussions/3818#discussioncomment-2155340 will be the magic
        // also no idea how to fetch privacy level with the below, however I guess all below are public as it is a public api endpoint
        const profiles = [];
        let hasNextPage = true, afterCursor = opts.afterCursor;
        while (hasNextPage) {
            // do not apply size default, as that causes shared opts to have pagination which isn't desired
            const filters = [`first: ${opts.size || 100}`];
            if (afterCursor)
                filters.push(`after: "${afterCursor}"`);
            const responseData = await queryGraphQL(
            /* GraphQL */ `{
					user(login: "${username}") {
						sponsors(${filters.join(', ')}) {
							pageInfo {
								hasNextPage
								endCursor
							}
							nodes {
								... on User {
									bio
									company
									email
									isHireable
									location
									login
									name
									url
									websiteUrl
								}
								... on Organization {
									description
									email
									location
									login
									name
									url
									websiteUrl
								}
							}
						}
					}
				}`, opts);
            const nodes = responseData.user?.sponsors?.nodes;
            if (!Array.isArray(nodes))
                throw new Error(`Response did not include an array of GitHub Sponsors for username: ${username}`);
            (0, list_1.append)(profiles, nodes);
            const pageInfo = responseData.user?.sponsors?.pageInfo || {
                hasNextPage: false,
                endCursor: '',
            };
            hasNextPage = pageInfo.hasNextPage;
            afterCursor = pageInfo.endCursor;
        }
        // convert
        const sponsors = fellow_1.default.add(profiles.map((githubProfile) => fellow_1.default.ensure({
            githubProfile,
            company: githubProfile.company,
            description: githubProfile.bio || githubProfile.description,
            email: githubProfile.email,
            githubUrl: githubProfile.url,
            githubUsername: githubProfile.login,
            hireable: githubProfile.isHireable,
            location: githubProfile.location,
            name: githubProfile.name,
            websiteUrl: githubProfile.websiteUrl,
        })));
        return { sponsors, donors: sponsors };
    }
    catch (err) {
        throw new errlop_1.default(`Fetching GitHub Sponsors failed for username: ${username}`, err);
    }
}
exports.getBackersFromGitHubSponsors = getBackersFromGitHubSponsors;
/** Get the {@link Fellow} entities for GitHub Contributors (minus bots) of this GitHub Repository via the GitHub REST API */
async function getGitHubContributors(githubSlug, opts = {}) {
    // fetch
    // https://docs.github.com/en/rest/reference/repos#list-repository-contributors
    const data = await queryREST({
        ...opts,
        pathname: `repos/${githubSlug}/contributors`,
        pages: opts.pages ?? 0,
    });
    if (data.length === 0)
        return [];
    // prepare
    const results = new Set();
    // add these items
    (0, list_1.append)(results, await Promise.all(data
        .filter((profile) => profile.login.includes('[bot]') === false)
        .map(async (profile) => {
        const fellow = await getGitHubProfileFromApiUrl(profile.url, opts);
        fellow.contributionsOfRepository.set(githubSlug, profile.contributions);
        fellow.contributorOfRepositories.add(githubSlug);
        return fellow;
    })));
    // return it all
    return fellow_1.default.sort(results);
}
exports.getGitHubContributors = getGitHubContributors;
/** Get the {@link Fellow} entities for GitHub Members of this GitHub Organisation via the GitHub REST API */
async function getGitHubMembersFromOrganization(org, opts = {}) {
    // fetch
    // https://docs.github.com/en/rest/reference/orgs#list-public-organization-members
    const data = await queryREST({
        ...opts,
        pathname: `orgs/${org}/public_members`,
    });
    if (data.length === 0)
        return [];
    // prepare
    const members = new Set();
    // add these items
    (0, list_1.append)(members, await Promise.all(data.map((profile) => getGitHubProfileFromApiUrl(profile.url, opts))));
    // return it all
    return fellow_1.default.sort(members);
}
exports.getGitHubMembersFromOrganization = getGitHubMembersFromOrganization;
/** Get the {@link Fellow} entities for the GitHub Members of these GitHub Organisations via the GitHub REST API */
async function getGitHubMembersFromOrganizations(orgs, opts = {}) {
    const lists = await Promise.all(orgs.map((org) => getGitHubMembersFromOrganization(org, opts)));
    return (0, list_1.flatten)(lists);
}
exports.getGitHubMembersFromOrganizations = getGitHubMembersFromOrganizations;
/** For each backer, append to the result */
function appendBackers(target, source) {
    for (const field of exports.backerFields) {
        const list = source[field];
        if (list)
            (0, list_1.append)(target[field], list);
    }
    return target;
}
/** For each backer, attach them to the slug for the appropriate associations */
function attachBackersToGitHubSlug(result, slug) {
    if (slug) {
        for (const author of result.author)
            author.authorOfRepositories.add(slug);
        for (const author of result.authors)
            author.authorOfRepositories.add(slug);
        for (const maintainer of result.maintainers)
            maintainer.maintainerOfRepositories.add(slug);
        for (const contributor of result.contributors)
            contributor.contributorOfRepositories.add(slug);
        for (const funder of result.funders)
            funder.funderOfRepositories.add(slug);
        for (const sponsor of result.sponsors)
            sponsor.sponsorOfRepositories.add(slug);
        for (const donor of result.donors)
            donor.donorOfRepositories.add(slug);
    }
    return result;
}
/** Verify URLs of backers to ensure we aren't putting bad data */
async function verifyUrlsOfBackers(source) {
    for (const field of exports.backerFields) {
        const list = source[field];
        if (list)
            await Promise.all(list.map((fellow) => fellow.verifyUrls()));
    }
    return source;
}
/** Get the {@link Backers} of this GitHub Repository */
async function getBackers(opts = {}) {
    let githubSlug = (typeof opts.githubSlug === 'string' && opts.githubSlug) || '', packageData = (opts.packageData &&
        typeof opts.packageData === 'object' &&
        opts.packageData) ||
        {};
    try {
        // if auto-detect package data, fetch from slug if not offline
        if (githubSlug && opts.packageData !== false && opts.offline !== true) {
            packageData = await getPackageData(githubSlug, {});
        }
        // if auto-detect slug, fetch from package data
        if (opts.githubSlug !== false && packageData) {
            githubSlug = getGitHubSlugFromPackageData(packageData) || '';
        }
        // prepare
        const authed = hasCredentials(opts.credentials);
        if (!authed && opts.offline !== true) {
            console.warn('GitHub credentials not provided, will skip fetching GitHub Contributors, Sponsors, and Profiles');
        }
        let fundingData;
        const fallbackUsername = githubSlug && githubSlug.split('/')[0];
        // get badge only usernames
        const thanksdevGithubUsername = opts.thanksdevGithubUsername === false
            ? ''
            : (typeof opts.thanksdevGithubUsername === 'string' &&
                opts.thanksdevGithubUsername) ||
                packageData.badges?.config?.thanksdevGithubUsername ||
                fallbackUsername;
        // get badge or funding usernames
        let githubSponsorsUsername = opts.githubSponsorsUsername === false
            ? ''
            : (typeof opts.githubSponsorsUsername === 'string' &&
                opts.githubSponsorsUsername) ||
                packageData.badges?.config?.githubSponsorsUsername ||
                '';
        let opencollectiveUsername = opts.opencollectiveUsername === false
            ? ''
            : (typeof opts.opencollectiveUsername === 'string' &&
                opts.opencollectiveUsername) ||
                packageData.badges?.config?.opencollectiveUsername ||
                '';
        if (opts.offline !== true &&
            (!githubSponsorsUsername || !opencollectiveUsername)) {
            fundingData = await getFundingData(githubSlug, {});
            if (!githubSponsorsUsername) {
                if (typeof fundingData.github === 'string') {
                    githubSponsorsUsername = fundingData.github || '';
                }
                else if (Array.isArray(fundingData.github) &&
                    fundingData.github.length >= 1) {
                    githubSponsorsUsername = fundingData.github[0] || '';
                }
            }
            if (!opencollectiveUsername) {
                if (typeof fundingData.open_collective === 'string') {
                    opencollectiveUsername = fundingData.open_collective || '';
                }
                else if (Array.isArray(fundingData.open_collective) &&
                    fundingData.open_collective.length >= 1) {
                    opencollectiveUsername = fundingData.open_collective[0] || '';
                }
            }
        }
        // shall we refresh sponsors? that is disregard existing package.json entries for remote data?
        const refreshSponsors = githubSponsorsUsername && authed && opts.offline !== true;
        // prepare backers from package data
        const result = attachBackersToGitHubSlug({
            author: fellow_1.default.add(packageData.author),
            authors: fellow_1.default.add(packageData.author, packageData.authors),
            maintainers: fellow_1.default.add(packageData.maintainers),
            contributors: fellow_1.default.add(packageData.maintainers, packageData.contributors),
            funders: fellow_1.default.add(packageData.funders),
            sponsors: refreshSponsors ? [] : fellow_1.default.add(packageData.sponsors),
            donors: fellow_1.default.add(packageData.funders, packageData.sponsors, packageData.donors),
        }, githubSlug);
        // if not offline, fetch from remote sources
        if (opts.offline !== true) {
            // ThanksDev
            if (thanksdevGithubUsername) {
                try {
                    const fetchedBackers = await getBackersFromThanksDev(ThanksDevPlatform.GitHub, thanksdevGithubUsername, opts);
                    appendBackers(result, fetchedBackers);
                }
                catch (err) {
                    console.warn(err.stack);
                }
            }
            else {
                console.warn(`Unable to fetch backers from ThanksDev, as unable to resolve the ThanksDev username for ${githubSlug}`);
            }
            // OpenCollective
            if (opencollectiveUsername) {
                try {
                    const fetchedBackers = await getBackersFromOpenCollective(opencollectiveUsername, opts);
                    appendBackers(result, fetchedBackers);
                }
                catch (err) {
                    console.warn(err.stack);
                }
            }
            else {
                console.warn(`Unable to fetch backers from OpenCollective, as unable to resolve the OpenCollective username for ${githubSlug}`);
            }
            // GitHub
            if (authed) {
                // GitHub Contributors
                try {
                    const fetchedContributors = await getGitHubContributors(githubSlug, opts);
                    (0, list_1.append)(result.contributors, fetchedContributors);
                }
                catch (err) {
                    console.warn(err.stack);
                }
                // GitHub Sponsors
                if (githubSponsorsUsername) {
                    try {
                        const fetchedBackers = await getBackersFromGitHubSponsors(githubSponsorsUsername, opts);
                        appendBackers(result, fetchedBackers);
                    }
                    catch (err) {
                        console.warn(err.stack);
                    }
                }
                else {
                    console.warn(`Unable to fetch backers from GitHub Sponsors, as unable to resolve the GitHub Sponsors username for ${githubSlug}`);
                }
                // fetch additional details, if able
                await Promise.all(Array.from(result.donors).map(async (fellow) => {
                    if (fellow.githubUsername && !fellow.githubProfile) {
                        await getGitHubProfile(fellow.githubUsername, opts);
                    }
                }));
            }
            // verify their details
            verifyUrlsOfBackers(result);
        }
        // attach to slug, which enables us to remove duplicates by fetching by slug
        // (duplicates can occur if new details allowed us to merge two old entries)
        attachBackersToGitHubSlug(result, githubSlug);
        return {
            author: result.author,
            authors: fellow_1.default.authorsOfRepository(githubSlug),
            maintainers: fellow_1.default.maintainersOfRepository(githubSlug),
            contributors: fellow_1.default.contributorsOfRepository(githubSlug),
            funders: fellow_1.default.fundersOfRepository(githubSlug),
            sponsors: fellow_1.default.sponsorsOfRepository(githubSlug),
            donors: fellow_1.default.donorsOfRepository(githubSlug),
        };
    }
    catch (err) {
        const error = new errlop_1.default(githubSlug
            ? `Failed to fetch backers for GitHub Repository: ${githubSlug}`
            : `Failed to fetch backers for: ${JSON.stringify(opts, null, '  ')}`, err);
        console.warn(error.stack);
        return {
            author: [],
            authors: [],
            maintainers: [],
            contributors: [],
            funders: [],
            sponsors: [],
            donors: [],
        };
    }
}
exports.getBackers = getBackers;
/** Get the {@link Backers} of these GitHub Repositories */
async function getBackersFromRepositories(githubSlugs, opts = {}) {
    const results = await Promise.all(githubSlugs.map((githubSlug) => getBackers({ ...opts, githubSlug })));
    // flattern and de-duplicate across repos
    const author = new Set();
    const authors = new Set();
    const maintainers = new Set();
    const contributors = new Set();
    const funders = new Set();
    const sponsors = new Set();
    const donors = new Set();
    for (const result of results) {
        (0, list_1.append)(author, result.author);
        (0, list_1.append)(authors, result.authors);
        (0, list_1.append)(maintainers, result.maintainers);
        (0, list_1.append)(contributors, result.contributors);
        (0, list_1.append)(funders, result.funders);
        (0, list_1.append)(sponsors, result.sponsors);
        (0, list_1.append)(donors, result.donors);
    }
    return {
        author: fellow_1.default.sort(author),
        authors: fellow_1.default.sort(authors),
        maintainers: fellow_1.default.sort(maintainers),
        contributors: fellow_1.default.sort(contributors),
        funders: fellow_1.default.sort(funders),
        sponsors: fellow_1.default.sort(sponsors),
        donors: fellow_1.default.sort(donors),
    };
}
exports.getBackersFromRepositories = getBackersFromRepositories;
/** Get the {@link Backers} for the GitHub Repositories of these usernames (users, organizations) */
async function getBackersFromUsernames(usernames, opts = {}) {
    const repos = await getGitHubRepositoriesFromUsernames(usernames, opts);
    // filter out forks and grab the slugs
    const slugs = repos
        .filter((repo) => repo.fork !== true)
        .map((repo) => repo.full_name);
    // Fetch the backers for the repos
    return getBackersFromRepositories(slugs, opts);
}
exports.getBackersFromUsernames = getBackersFromUsernames;
/**
 * Get the {@link Backers} for the GitHub Repositories of this search query
 * @param query e.g.`\@bevry language:typescript`
 */
async function getBackersFromSearch(query, opts = {}) {
    const repos = await getGitHubRepositoriesFromSearch(query, opts);
    // Just grab the slugs
    const githubSlugs = repos.map((repo) => repo.full_name);
    // Fetch the backers for the repos
    return getBackersFromRepositories(githubSlugs, opts);
}
exports.getBackersFromSearch = getBackersFromSearch;
/** Render the {@link Backers} */
function renderBackers(backers, opts) {
    // prepare
    const githubSlug = opts.githubSlug === false
        ? ''
        : (typeof opts.githubSlug === 'string' && opts.githubSlug) ||
            (opts.packageData && getGitHubSlugFromPackageData(opts.packageData)) ||
            '';
    const projectName = opts.projectName === false
        ? ''
        : (typeof opts.projectName === 'string' && opts.projectName) ||
            opts.packageData?.title ||
            opts.packageData?.name ||
            '';
    // render
    if (opts.format === BackersRenderFormat.string ||
        opts.format === BackersRenderFormat.text ||
        opts.format === BackersRenderFormat.markdown ||
        opts.format === BackersRenderFormat.html) {
        const author = backers.author?.map((fellow) => fellow.toFormat({
            displayYears: true,
            ...opts,
            displayDescription: false, // override
            githubSlug,
            format: opts.format,
        })) || null;
        return (0, trim_empty_keys_1.default)({
            author: opts.format === BackersRenderFormat.string
                ? author?.join(', ')
                : author,
            authors: backers.authors?.map((fellow) => fellow.toFormat({
                displayYears: true,
                displayDescription: true,
                ...opts,
                githubSlug,
                format: opts.format,
            })) || null,
            maintainers: backers.maintainers?.map((fellow) => fellow.toFormat({
                displayDescription: true,
                ...opts,
                githubSlug,
                format: opts.format,
            })) || null,
            contributors: backers.contributors?.map((fellow) => fellow.toFormat({
                // show contributions as such use github url
                displayContributions: true,
                urlFields: opts.format === BackersRenderFormat.string
                    ? []
                    : ['githubUrl', 'url'],
                ...opts,
                githubSlug,
                format: opts.format,
            })) || null,
            funders: backers.funders?.map((fellow) => fellow.toFormat({
                displayDescription: true,
                ...opts,
                githubSlug,
                format: opts.format,
            })) || null,
            sponsors: backers.sponsors?.map((fellow) => fellow.toFormat({
                displayDescription: true,
                ...opts,
                githubSlug,
                format: opts.format,
            })) || null,
            donors: backers.donors?.map((fellow) => fellow.toFormat({
                ...opts,
                githubSlug,
                format: opts.format,
            })) || null,
        });
    }
    else if (opts.format === BackersRenderFormat.package) {
        return (0, trim_empty_keys_1.default)((0, arrange_package_json_1.default)({
            ...(opts.packageData || {}),
            ...renderBackers(backers, {
                ...opts,
                githubSlug,
                format: BackersRenderFormat.string,
            }),
        }));
    }
    else if (opts.format === BackersRenderFormat.copyright) {
        return (0, trim_empty_keys_1.default)({
            author: backers.author?.map((fellow) => fellow.toMarkdown({
                displayCopyright: true,
                displayYears: true,
                ...opts,
                githubSlug,
                format: fellow_1.Format.html,
            })) || null,
            authors: backers.authors?.map((fellow) => fellow.toMarkdown({
                displayCopyright: true,
                displayYears: true,
                ...opts,
                githubSlug,
                format: fellow_1.Format.html,
            })) || null,
        });
    }
    else if (opts.format === BackersRenderFormat.shoutout) {
        const projectNameAndSpace = projectName ? projectName + ' ' : '';
        return [
            ...(backers.contributors || []).map((fellow) => `Thank you to ${projectNameAndSpace}contributor 🤍 ${fellow.toText({
                ...opts,
                githubSlug,
                format: fellow_1.Format.text,
            })}`),
            ...(backers.funders || []).map((fellow) => `Thank you to ${projectNameAndSpace}funder 🤍 ${fellow.toText({
                ...opts,
                githubSlug,
                format: fellow_1.Format.text,
            })}`),
            ...(backers.sponsors || []).map((fellow) => `Thank you to ${projectNameAndSpace}sponsor 🤍 ${fellow.toText({
                ...opts,
                githubSlug,
                format: fellow_1.Format.text,
            })}`),
        ].join('\n');
    }
    else if (opts.format === BackersRenderFormat.update) {
        return (backers.sponsors || []).length === 0
            ? ''
            : ('- Thank you to the sponsors: ' +
                backers.sponsors
                    ?.map((fellow) => fellow.toMarkdown({
                    ...opts,
                    githubSlug,
                    format: fellow_1.Format.markdown,
                }))
                    .join(', '));
    }
    else if (opts.format === BackersRenderFormat.release) {
        return [
            (backers.funders || []).length === 0
                ? ''
                : '- Thank you to the funders: ' +
                    backers.funders
                        ?.map((fellow) => fellow.toMarkdown({
                        ...opts,
                        githubSlug,
                        format: fellow_1.Format.markdown,
                    }))
                        .join(', '),
            (backers.sponsors || []).length === 0
                ? ''
                : '- Thank you to the sponsors: ' +
                    backers.sponsors
                        ?.map((fellow) => fellow.toMarkdown({
                        ...opts,
                        githubSlug,
                        format: fellow_1.Format.markdown,
                    }))
                        .join(', '),
        ]
            .filter(Boolean)
            .join('\n');
    }
    else {
        throw new Error(`Invalid format: ${opts.format}`);
    }
}
exports.renderBackers = renderBackers;
